//***************************************************************************
//* Copyright (c) 2018-2019 Saint Petersburg State University
//* All Rights Reserved
//* See file LICENSE for details.
//***************************************************************************

#include "io/utils/edge_namer.hpp"
#include "io/binary/graph.hpp"
#include "io/reads/io_helper.hpp"
#include "io/reads/wrapper_collection.hpp"
#include "io/reads/multifile_reader.hpp"
#include "io/reads/file_reader.hpp"
#include "io/graph/gfa_reader.hpp"
#include "io/graph/gfa_writer.hpp"
#include "assembly_graph/core/graph.hpp"
#include "utils/logger/log_writers.hpp"
#include "modules/alignment/pacbio/g_aligner.hpp"

#include "mapping_printer.hpp"
#include "spaligner_config.hpp"

#include "llvm/Support/YAMLParser.h"
#include "llvm/Support/YAMLTraits.h"

#include <iostream>
#include <fstream>
#include <clipp/clipp.h>

using namespace std;

void create_console_logger() {
    logging::logger *log = logging::create_logger("", logging::L_INFO);
    log->add_writer(make_shared<logging::console_cerr_writer>());
    logging::attach_logger(log);
}

namespace llvm {
namespace yaml {

template<> struct MappingTraits<debruijn_graph::config::pacbio_processor> {
    static void mapping(IO& io, debruijn_graph::config::pacbio_processor& cfg) {
        io.mapRequired("internal_length_cutoff", cfg.internal_length_cutoff);
        io.mapRequired("path_limit_stretching", cfg.path_limit_stretching);
        io.mapRequired("path_limit_pressing", cfg.path_limit_pressing);
        io.mapRequired("max_path_in_chaining", cfg.max_path_in_dijkstra);
        io.mapRequired("max_vertex_in_chaining", cfg.max_vertex_in_dijkstra);
    }
};

template<> struct MappingTraits<sensitive_aligner::GapClosingConfig> {
    static void mapping(IO& io, sensitive_aligner::GapClosingConfig& cfg) {
        io.mapRequired("queue_limit", cfg.queue_limit);
        io.mapRequired("iteration_limit", cfg.iteration_limit);
        io.mapRequired("updates_limit", cfg.updates_limit);
        io.mapRequired("find_shortest_path", cfg.find_shortest_path);
        io.mapRequired("restore_mapping", cfg.restore_mapping);
        io.mapRequired("penalty_ratio", cfg.penalty_ratio);
        io.mapRequired("max_ed_proportion", cfg.max_ed_proportion);
        io.mapRequired("ed_lower_bound", cfg.ed_lower_bound);
        io.mapRequired("ed_upper_bound", cfg.ed_upper_bound);
        io.mapRequired("max_gs_states", cfg.max_gs_states);
    }
};

template<> struct MappingTraits<sensitive_aligner::EndsClosingConfig> {
    static void mapping(IO& io, sensitive_aligner::EndsClosingConfig& cfg) {
        io.mapRequired("queue_limit", cfg.queue_limit);
        io.mapRequired("iteration_limit", cfg.iteration_limit);
        io.mapRequired("updates_limit", cfg.updates_limit);
        io.mapRequired("find_shortest_path", cfg.find_shortest_path);
        io.mapRequired("restore_mapping", cfg.restore_mapping);
        io.mapRequired("penalty_ratio", cfg.penalty_ratio);
        io.mapRequired("max_ed_proportion", cfg.max_ed_proportion);
        io.mapRequired("ed_lower_bound", cfg.ed_lower_bound);
        io.mapRequired("ed_upper_bound", cfg.ed_upper_bound);
        io.mapRequired("max_restorable_length", cfg.max_restorable_length);
    }
};

template<> struct MappingTraits<sensitive_aligner::ProteinAlignmentConfig> {
    static void mapping(IO& io, sensitive_aligner::ProteinAlignmentConfig& cfg) {
        io.mapRequired("queue_limit", cfg.queue_limit);
        io.mapRequired("iteration_limit", cfg.iteration_limit);
        io.mapRequired("updates_limit", cfg.updates_limit);
        io.mapRequired("find_shortest_path", cfg.find_shortest_path);
        io.mapRequired("restore_mapping", cfg.restore_mapping);
        io.mapRequired("max_ed_proportion", cfg.max_ed_proportion);
        io.mapRequired("max_restorable_length", cfg.max_restorable_length);
        io.mapRequired("stop_codon", cfg.stop_codon);
        io.mapRequired("min_alignment_len", cfg.min_alignment_len);
        io.mapRequired("penalty_matrix", cfg.penalty_matrix);
        io.mapRequired("indel_score", cfg.indel_score);
    }
};

template<> struct MappingTraits<sensitive_aligner::GAlignerConfig> {
    static void mapping(IO& io, sensitive_aligner::GAlignerConfig& cfg) {
        io.mapRequired("run_dijkstra", cfg.hit_extension_cfg.gap_cfg.run_dijkstra);
        io.mapRequired("restore_ends", cfg.restore_ends);

        io.mapRequired("hits_generation", cfg.pb);
        io.mapRequired("gap_closing", cfg.hit_extension_cfg.gap_cfg);
        io.mapRequired("ends_recovering", cfg.hit_extension_cfg.ends_cfg);
        io.mapRequired("aminoacid_mode", cfg.hit_extension_cfg.protein_cfg);
    }
};

}
}

namespace sensitive_aligner {

class LongReadsAligner {
  public:
    LongReadsAligner(const debruijn_graph::ConjugateDeBruijnGraph &g,
                     const io::CanonicalEdgeHelper<debruijn_graph::Graph> &edge_namer,
                     const GAlignerConfig &cfg,
                     const int threads)
        :g_(g),
         cfg_(cfg),
         galigner_(g_, cfg.data_type, cfg.pb, cfg.hit_extension_cfg, cfg.restore_ends),
         threads_(threads),
         mapping_printer_hub_(g_, cfg.data_type, edge_namer, cfg.output_file, cfg.output_fasta, cfg.output_proteinfasta) {
        aligned_reads_ = 0;
        processed_reads_ = 0;
    }

    void RunAligner() {
        auto read_stream = io::FileReadStream(cfg_.path_to_sequences);
        if (alignment::BWAIndex::AlignmentMode::Protein != cfg_.data_type) {
            auto read_stream = io::FixingWrapper(io::FileReadStream(cfg_.path_to_sequences));
        }
        size_t n = 0;
        size_t buffer_no = 0;
        while (!read_stream.eof()) {
            std::vector<io::SingleRead> read_buffer;
            read_buffer.reserve(read_buffer_size);
            io::SingleRead read;
            for (size_t buf_size = 0; buf_size < read_buffer_size && !read_stream.eof(); ++buf_size) {
                read_stream >> read;
                if (alignment::BWAIndex::AlignmentMode::Protein == cfg_.data_type) {
                    io::SingleRead read_nuc;
                    read_nuc = io::SingleRead(read.name(), ConvertProtein2CanonicalNuc(read.GetSequenceString()));
                    read = read_nuc;
                }
                read_buffer.push_back(move(read));
            }
            INFO("Prepared batch " << buffer_no << " of " << read_buffer.size() << " sequences");
            AlignBatch(read_buffer);
            ++buffer_no;
            n += read_buffer.size();
            INFO("Processed " << n << " sequences");
        }
    }

  private:

    OneReadMapping AlignRead(const io::SingleRead &read) const {
        DEBUG("Sequence " << read.name() << ". Current Sequence")
        utils::perf_counter pc;
        auto current_read_mapping = galigner_.GetAlignment(read);
        const auto& aligned_mappings = current_read_mapping.edge_paths;
        if (aligned_mappings.size() > 0) {
            DEBUG("Sequence " << read.name() << " is aligned");
        }  else {
            DEBUG("Sequence " << read.name() << " wasn't aligned and length=" << read.sequence().size());
        }
        DEBUG("Sequence " << read.name() << " seq_time=" << pc.time())
        return current_read_mapping;
    }

    void AlignBatch(const vector<io::SingleRead> &reads) {
        size_t step = 10;
        processed_reads_ = 0;
        aligned_reads_ = 0;
        #pragma omp parallel for schedule(guided, 50) num_threads(threads_)
        for (size_t i = 0 ; i < reads.size(); ++i) {
            OneReadMapping res = AlignRead(reads[i]);
            if (res.edge_paths.size() > 0) {
                mapping_printer_hub_.SaveMapping(res, reads[i]);
            }
            #pragma omp critical(aligner)
            {
                if (res.edge_paths.size() > 0) {
                    aligned_reads_ ++;
                }
                processed_reads_ ++;
                if (processed_reads_ * 100 / reads.size() >= step) {
                    INFO("Processed \% sequences: " << processed_reads_ * 100 / reads.size() <<
                         "\%, Aligned sequences: " << aligned_reads_ * 100 / processed_reads_ <<
                         "\% (" << aligned_reads_ << " out of " << processed_reads_ << ")")
                    step += 10;
                }
            }
        }
    }

    const size_t read_buffer_size = 50000;

    const debruijn_graph::ConjugateDeBruijnGraph &g_;
    const GAlignerConfig &cfg_;
    const sensitive_aligner::GAligner galigner_;
    const int threads_;
    MappingPrinterHub mapping_printer_hub_;

    int aligned_reads_;
    int processed_reads_;

};

void Launch(GAlignerConfig &cfg, int threads) {
    string tmpdir = fs::make_temp_dir(fs::current_dir(), "tmp");

    DEBUG("Load gfa");
    VERIFY_MSG(fs::is_regular_file(cfg.path_to_graphfile), "GFA-file " + cfg.path_to_graphfile + " doesn't exist");
    io::IdMapper<std::string> id_mapper;
    gfa::GFAReader gfa(cfg.path_to_graphfile);
    DEBUG("Segments: " << gfa.num_edges() << ", links: " << gfa.num_links());
    VERIFY_MSG(gfa.k() != -1U, "Failed to determine k-mer length");
    VERIFY_MSG(gfa.k() % 2 == 1, "k-mer length must be odd");
    debruijn_graph::ConjugateDeBruijnGraph g(gfa.k());
    cfg.K = gfa.k();
    gfa.to_graph(g, &id_mapper);

    io::CanonicalEdgeHelper<debruijn_graph::Graph> edge_namer(g, io::MapNamingF<debruijn_graph::Graph>(id_mapper));
    INFO("Loaded graph with " << g.size() << " vertices and k-mer size " << g.k());

    LongReadsAligner aligner(g, edge_namer, cfg, threads);
    INFO("LongSequenceAligner created");

    INFO("Process sequences from " << cfg.path_to_sequences);
    aligner.RunAligner();
    INFO("Thank you for using SPAligner!")
    fs::remove_dir(tmpdir);
}
} // namespace sensitive_aligner

void process_cmdline(int argc, char **argv,
                    sensitive_aligner::GAlignerConfig &config,
                    string &cfg_name,
                    string &seq_type,
                    unsigned &nthreads) {
    using namespace clipp;

    auto cli = (
      (required("-d","--datatype") & value("type", seq_type)
                                           .if_missing([]{ cout << "ERROR: Sequence type is not provided (nanopore or pacbio)\n"; } ))
                                           % "type of sequences: nanopore, pacbio, protein",
      (required("-s","--sequences") & value("fasta", config.path_to_sequences)
                                           .if_missing([]{ cout << "ERROR: Path to file with sequences is not provided\n"; } ))
                                           % "path to fasta/fastq file with sequences (gzip archives supported)",
      (required("-g","--graph") & value("graph", config.path_to_graphfile)
                                        .if_missing([]{ cout << "ERROR: Path to file with graph is not provided\n"; } ))
                                         % "path to GFA1-file",
      (option("-t", "--threads") & integer("value", nthreads)) % "# of threads to use",
      (option("-c", "--config") & value("cfg", cfg_name)) % "config with SPAligner parameters description (in YAML)",
      (option("-f", "--fasta")  & value("filename", config.output_fasta)) % "file to output FASTA with nucleotide sequence alignments",
      (option("-p", "--proteinfasta") & value("filename", config.output_proteinfasta)) % "file to output FASTA with amino acid sequence alignments (only for protein mode)",
      (option("-o", "--output") & value("filename", config.output_file)) % "TSV output file (default stdout)"
    );

    auto result = parse(argc, argv, cli);
    if (!result) {
      std::cout << make_man_page(cli, argv[0]);
      exit(1);
    }
    
    if (seq_type == "nanopore")
        config.data_type = alignment::BWAIndex::AlignmentMode::Ont2D;
    else if (seq_type == "pacbio")
        config.data_type = alignment::BWAIndex::AlignmentMode::PacBio;
    else if (seq_type == "protein")
        config.data_type = alignment::BWAIndex::AlignmentMode::Protein;
    else {
        cerr << "You need to provide a supported datatype - nanopore, pacbio or protein" << endl;
        std::cout << make_man_page(cli, argv[0]);
        exit(-1);
    }
}

int main(int argc, char **argv) {

    unsigned nthreads = 8;
    string cfg, seq_type;
    sensitive_aligner::GAlignerConfig config;

    process_cmdline(argc, argv, config, cfg, seq_type, nthreads);

    std::string cmd_line = "";
    for (int i = 0; i < argc; ++ i) {
        cmd_line += std::string(argv[i]) + " ";
    }

    create_console_logger();
    START_BANNER("SPAligner: long sequence to graph alignment");
    INFO("Command line: " << cmd_line);
    if (config.output_proteinfasta != "" && seq_type != "protein") {
        WARN("File with amino acid sequence won't be generated as datatype is not set to protein")
        config.output_proteinfasta = "";
    }
    if (cfg.size() > 0) {
        INFO("Loading config from " << cfg)
        auto buf = llvm::MemoryBuffer::getFile(cfg);
        VERIFY_MSG(buf, "Failed to load config file " + cfg);
        llvm::yaml::Input yin(*buf.get());
        yin >> config;
        omp_set_num_threads(nthreads);
    } else {
        INFO("Running SPAligner with default parameters. No config was loaded.")
        config.pb.internal_length_cutoff = 200;
        config.pb.path_limit_stretching = 1.3;
        config.pb.path_limit_pressing = 0.6;
        config.pb.max_path_in_dijkstra = 15000;
        config.pb.max_vertex_in_dijkstra = 5000;

        config.hit_extension_cfg.ends_cfg.find_shortest_path = true;
        config.hit_extension_cfg.ends_cfg.penalty_ratio = float(0.1);
        config.hit_extension_cfg.ends_cfg.max_ed_proportion = 5;
    }

    sensitive_aligner::Launch(config, nthreads);
    return 0;
}
